unit hiAnalogGauge;

interface

uses Windows,Messages,Kol,Share,Win;

const
  MaxPixelCount = MaxInt div SizeOf(TRGBTriple);

type
  PRGBArray = ^TRGBArray;
  TRGBArray = array[0..MaxPixelCount-1] of TRGBTriple;

type
  PColor = ^TColor;

type
   TScale      =  (scPlus, scMinus, scMinusZPlus);
   TStyle      =  (LeftStyle, RightStyle, CenterStyle);
   TFaceOption =  (ShowMargin, ShowCircles, ShowMainTicks, ShowSubTicks,
                   ShowIndicatorMin, ShowIndicatorMid, ShowIndicatorMax,
                   ShowValues, ShowCenter, ShowFrame, Show3D, ShowCaption,ShowLed);
   TFaceOptions = set of TFaceOption;
   TAntialiased = (aaNone, aaBiline, aaTriline, aaQuadral);

type

 THIAnalogGauge = class(THIWin)

  private
    // face elements colors
    FMinColor:     TColor;
    FMidColor:     TColor;
    FMaxColor:     TColor;
    FFaceColor:    TColor;
    FColorFace:    TColor;
    FFonHover:     TColor;
    FTicksColor:   TColor;
    FValueColor:   TColor;
    FCaptionColor: TColor;
    FArrowColor:   TColor;
    FMarginColor:  TColor;
    FCenterColor:  TColor;
    FCircleColor:  TColor;
    FLedMaxColor:  TColor;

    // face elements sizes, etc.
    FCenterRadius: Integer;
    FCircleRadius: Integer;
    FScaleAngle: Integer;
    FMargin: Integer;
    FStyle: TStyle;
    FArrowWidth: Integer;
    FNumMainTicks: Integer;
    FLengthMainTicks: Integer;
    FLengthSubTicks: Integer;
    FFaceOptions: TFaceOptions;
    FIndent: Integer;
    GFont:PGraphicTool;
    FTypeScale: TScale;


    // values
    FPosition: Real;
    FScaleValue: Integer;
    FMinimum: Integer;
    FMaximum: Integer;
    FCaption: string;
    FColorHover: boolean;
    FGaugeMirror: boolean;
        
    // anti-aliasing mode
    FAntiAliased: TAntialiased;

    // internal bitmaps
    FBackBitmap: PBitmap;
    FFaceBitmap: PBitmap;
    FAABitmap: PBitmap;

    // show mode
    FShowMargin,FShowCircles,FShowMainTicks,FShowSubTicks,FShowIndicatorMin,
    FShowIndicatorMid,FShowIndicatorMax,FShowValues,FShowCenter,
    FShowFrame,FShow3D,FShowCaption,FShowLed: boolean;

    function GetAAMultipler: Integer;
    procedure DrawScale(Bitmap: PBitmap; K: Integer);
    procedure DrawArrow(Bitmap: PBitmap; K: Integer);
    procedure RedrawScale;
    procedure RedrawArrow;
    procedure FastAntiAliasPicture;
    procedure SetFontRec(Value: TFontRec);
    procedure SetOption(OSet:boolean; Option:TFaceOption);
    procedure Reinitialize;
    procedure TFontToGFont(Font:TFontRec; grFont:PGraphicTool);

    procedure _OnClick( Sender: PObj );
    procedure _OnMouseEnter( Sender: PObj );
    procedure _OnMouseLeave( Sender: PObj );
  public

    _data_Poz : THI_Event;
    _event_OnClick : THI_Event;
 
    property _prop_MinColor:         TColor         write FMinColor;
    property _prop_MidColor:         TColor         write FMidColor;
    property _prop_MaxColor:         TColor         write FMaxColor;
    property _prop_FaceColor:        TColor         write FColorFace;
    property _prop_FonHover:         TColor         write FFonHover;
    property _prop_TicksColor:       TColor         write FTicksColor;
    property _prop_ValueColor:       TColor         write FValueColor;
    property _prop_CaptionColor:     TColor         write FCaptionColor;
    property _prop_ArrowColor:       TColor         write FArrowColor;
    property _prop_MarginColor:      TColor         write FMarginColor;
    property _prop_CenterColor:      TColor         write FCenterColor;
    property _prop_CircleColor:      TColor         write FCircleColor;
    property _prop_LedMaxColor:      TColor         write FLedMaxColor;

    property _prop_ShowMargin:       boolean        write FShowMargin; 
    property _prop_ShowCircles:      boolean        write FShowCircles;
    property _prop_ShowMainTicks:    boolean        write FShowMainTicks;
    property _prop_ShowSubTicks:     boolean        write FShowSubTicks;
    property _prop_ShowIndicatorMin: boolean        write FShowIndicatorMin;
    property _prop_ShowIndicatorMid: boolean        write FShowIndicatorMid;
    property _prop_ShowIndicatorMax: boolean        write FShowIndicatorMax;
    property _prop_ShowValues:       boolean        write FShowValues;
    property _prop_ShowCenter:       boolean        write FShowCenter;
    property _prop_ShowFrame:        boolean        write FShowFrame;
    property _prop_Show3D:           boolean        write FShow3D;
    property _prop_ShowCaption:      boolean        write FShowCaption;
    property _prop_ShowLed:          boolean        write FShowLed;

    property _prop_TypeScale:        TScale         write FTypeScale;
    property _prop_ColorHover:       boolean        write FColorHover;
    property _prop_GaugeIndent:      Integer        write FIndent;
    property _prop_CenterRadius:     Integer        write FCenterRadius;
    property _prop_CircleRadius:     Integer        write FCircleRadius;
    property _prop_Angle:            Integer        write FScaleAngle;
    property _prop_GaugeMargin:      Integer        write FMargin;
    property _prop_ArrowWidth:       Integer        write FArrowWidth;
    property _prop_NumberMainTicks:  Integer        write FNumMainTicks;
    property _prop_LengthMainTicks:  Integer        write FLengthMainTicks;
    property _prop_LengthSubTicks:   Integer        write FLengthSubTicks;
    property _prop_Scale:            Integer        write FScaleValue;
    property _prop_IndMaximum:       Integer        write FMaximum;
    property _prop_IndMinimum:       Integer        write FMinimum;
    property _prop_GaugeCaption:     string         write FCaption;
    property _prop_GaugeStyle:       TStyle         write FStyle;
    property _prop_GaugeMirror:      boolean        write FGaugeMirror;
    property _prop_AntiAliased:      TAntialiased   write FAntiAliased;

    procedure  Init;    override;
    destructor Destroy; override;  
  
    procedure _work_doPosition(var _Data:TData; Index:word);
    procedure _var_Position(var _Data:TData; Index:word);

    procedure _work_doGaugeMirror(var _Data:TData; Index:word);
    procedure _work_doGaugeCaption(var _Data:TData; Index:word);
    procedure _work_doGaugeMargin(var _Data:TData; Index:word);
    procedure _work_doAngle(var _Data:TData; Index:word);
    procedure _work_doCenterRadius(var _Data:TData; Index:word);
    procedure _work_doCircleRadius(var _Data:TData; Index:word);
    procedure _work_doNumberMainTicks(var _Data:TData; Index:word);
    procedure _work_doLengthMainTicks(var _Data:TData; Index:word);
    procedure _work_doLengthSubTicks(var _Data:TData; Index:word);
    procedure _work_doArrowWidth(var _Data:TData; Index:word);
    procedure _work_doIndMinimum(var _Data:TData; Index:word);
    procedure _work_doIndMaximum(var _Data:TData; Index:word);
    procedure _work_doScale(var _Data:TData; Index:word);
    procedure _work_doColorHover(var _Data:TData; Index:word);
    procedure _work_doAntiAliased(var _Data:TData; Index:word);
    procedure _work_doGaugeIndent(var _Data:TData; Index:word);
    procedure _work_doTypeScale(var _Data:TData; Index:word);    

    procedure _work_doShowMargin(var _Data:TData; Index:word);
    procedure _work_doShowCircles(var _Data:TData; Index:word);
    procedure _work_doShowMainTicks(var _Data:TData; Index:word);
    procedure _work_doShowSubTicks(var _Data:TData; Index:word);
    procedure _work_doShowIndicatorMin(var _Data:TData; Index:word);
    procedure _work_doShowIndicatorMid(var _Data:TData; Index:word);
    procedure _work_doShowIndicatorMax(var _Data:TData; Index:word);
    procedure _work_doShowValues(var _Data:TData; Index:word);
    procedure _work_doShowCenter(var _Data:TData; Index:word);
    procedure _work_doShowFrame(var _Data:TData; Index:word);
    procedure _work_doShow3D(var _Data:TData; Index:word);
    procedure _work_doShowCaption(var _Data:TData; Index:word);
    procedure _work_doShowLed(var _Data:TData; Index:word);

    procedure _work_doLedMaxColor(var _Data:TData; Index:word);
    procedure _work_doMinColor(var _Data:TData; Index:word);
    procedure _work_doMidColor(var _Data:TData; Index:word);
    procedure _work_doMaxColor(var _Data:TData; Index:word);
    procedure _work_doFaceColor(var _Data:TData; Index:word);
    procedure _work_doFonHover(var _Data:TData; Index:word);
    procedure _work_doCaptionColor(var _Data:TData; Index:word);
    procedure _work_doTicksColor(var _Data:TData; Index:word);
    procedure _work_doValueColor(var _Data:TData; Index:word);
    procedure _work_doCircleColor(var _Data:TData; Index:word);
    procedure _work_doArrowColor(var _Data:TData; Index:word);
    procedure _work_doCenterColor(var _Data:TData; Index:word);
    procedure _work_doMarginColor(var _Data:TData; Index:word);
    procedure _work_doFont(var _Data:TData; Index:word);

  end;

implementation

function GetShadeColor(Color: TColor; Shade: Byte) : TColor;
var   fFrom: TRGB;
begin
  PColor(@fFrom)^:= Color2RGB(Color);
  Result := RGB(
    Max(0, FFrom.R - Shade),
    Max(0, FFrom.G - Shade),
    Max(0, FFrom.B - Shade)
  );
end;

function GetLightColor(Color: TColor; Light: Byte) : TColor;
var   fFrom: TRGB;
begin
  PColor(@fFrom)^:= Color2RGB(Color);
  Result := RGB(
    (FFrom.R*100 + (255 - FFrom.R) * Light) div 100,
    (FFrom.G*100 + (255 - FFrom.G) * Light) div 100,
    (FFrom.B*100 + (255 - FFrom.B) * Light) div 100
  );
end;

// SinCos is 2x faster than calling Sin and Cos separately for the same angle 
procedure SinCos(Theta: Extended; var Sin, Cos: Extended);
asm
        FLD     Theta
        FSINCOS
        FSTP    tbyte ptr [edx]    // Cos
        FSTP    tbyte ptr [eax]    // Sin
        FWAIT
end;

function WndProcGauge( Sender: PControl; var Msg: TMsg; var Rslt: Integer ): Boolean;
var   P: TPaintSTruct;
      fControl: THIAnalogGauge; 
begin
   Result:=False;
   fControl := THIAnalogGauge(Sender.Tag); 
   if msg.message = WM_PAINT then begin  
      BeginPaint(Sender.Handle,P);
         with fControl do begin
            if FAntiAliased = aaNone then
               BitBlt(P.hdc, 0, 0, FFaceBitmap.Width, FFaceBitmap.Height, FFaceBitmap.Canvas.Handle, 0, 0, SRCCOPY)
            else
               BitBlt(P.hdc, 0, 0, FAABitmap.Width, FAABitmap.Height, FAABitmap.Canvas.Handle, 0, 0, SRCCOPY);
         end;
      EndPaint(Sender.Handle,P);
      Result := True;
      Rslt := 0;
   end else if msg.message = WM_SIZE then begin
        if fControl.fStyle in [LeftStyle, RightStyle] then begin
           Sender.Width:=Min(Sender.Width,Sender.Height);
           Sender.Height:=Min(Sender.Width,Sender.Height);
        end;
        fControl.ReInitialize;
   end;
end;

{ ========================================================================= }

procedure THIAnalogGauge.TFontToGFont;
begin
   GrFont.Color:= Font.Color;
   Share.SetFont(GrFont,Font.Style);
   GrFont.FontName:= Font.Name;
   GrFont.FontHeight:= _hi_SizeFnt(Font.Size);
   GrFont.FontCharset:= Font.CharSet;
end;

procedure THIAnalogGauge.Init;
begin
   Control:= NewPaintBox(FParent);
   Control.Tag := longint(Self);

   SetFontRec(_prop_Font);
   FFaceColor := FColorFace;

   FFaceOptions:= [ ];
   if FShowMargin       then include(FFaceOptions,ShowMargin);
   if FShowCircles      then include(FFaceOptions,ShowCircles);
   if FShowMainTicks    then include(FFaceOptions,ShowMainTicks);
   if FShowSubTicks     then include(FFaceOptions,ShowSubTicks);
   if FShowIndicatorMin then include(FFaceOptions,ShowIndicatorMin);
   if FShowIndicatorMid then include(FFaceOptions,ShowIndicatorMid);
   if FShowIndicatorMax then include(FFaceOptions,ShowIndicatorMax);
   if FShowValues       then include(FFaceOptions,ShowValues);
   if FShowCenter       then include(FFaceOptions,ShowCenter);
   if FShowFrame        then include(FFaceOptions,ShowFrame);
   if FShow3D           then include(FFaceOptions,Show3D);
   if FShowCaption      then include(FFaceOptions,ShowCaption);
   if FShowLed          then include(FFaceOptions,ShowLed);

   FPosition := 0;

   if FScaleValue <= FMaximum then FScaleValue := FMaximum + 1;
   
   if (FMinimum < 1) or (FMinimum > FScaleValue) then FMinimum := 1;
   if FMinimum = FScaleValue then FMinimum := FScaleValue - 1;
   if FMinimum >= FMaximum then FMinimum := FMaximum - 1;

   if (FMaximum < 1) or (FMaximum > FScaleValue) then FMaximum := 1;
   if FMaximum = FScaleValue then FMaximum := FScaleValue - 1;
   if FMaximum <= FMinimum then FMaximum := FMinimum + 1;    
  
   if (FScaleAngle < 10) and (FScaleAngle > 360) then FScaleAngle := 10;
   if FArrowWidth < 1 then FArrowWidth := 1 else if FArrowWidth > 10 then FArrowWidth := 10;
   if FNumMainTicks < 2 then FNumMainTicks := 2;

   Control.AttachProc(WndProcGauge);

   Control.OnClick := _OnClick;
   Control.OnMouseEnter :=_OnMouseEnter;
   Control.OnMouseLeave :=_OnMouseLeave;
inherited;
   RedrawScale;
end;

destructor THIAnalogGauge.Destroy;
begin
  if Assigned(FFaceBitmap) then FFaceBitmap.Free;
  if Assigned(FBackBitmap) then FBackBitmap.Free;
  if Assigned(FaaBitmap)   then FaaBitmap.Free;
  if Assigned(GFont) then GFont.free;
  inherited;
end;

{ ------------------------------------------------------------------------- }

procedure THIAnalogGauge._work_doPosition;
begin
  FPosition := ReadReal(_Data, _data_Poz);
  RedrawArrow;
  Control.Invalidate;
end;

procedure THIAnalogGauge._var_Position;
begin
   dtReal(_Data, FPosition);
end;

procedure THIAnalogGauge._OnClick;
begin
   _hi_OnEvent(_event_OnClick);
end;
  
procedure THIAnalogGauge._OnMouseEnter;
begin
   if not FColorHover then exit;
   FFaceColor := FFonHover;
   RedrawScale;
end;

procedure THIAnalogGauge._OnMouseLeave;
begin
   if not FColorHover then exit;
   FFaceColor := FColorFace;
   RedrawScale;
end;

procedure THIAnalogGauge.SetFontRec;
begin
  if Assigned(GFont) then GFont.free;
  GFont := NewFont;
  TFontToGFont(Value, GFont);
  Control.Font.Assign(GFont);
end;

{ ------------------------------------------------------------------------- }

procedure THIAnalogGauge.Reinitialize;
var   K: Integer;
begin
   with Control{$ifndef F_P}^{$endif} do begin
      if Width < 30 then Width :=30;
      if Height < 30 then Height:=30;
      K := GetAAMultipler;
      if Assigned(FFaceBitmap) then FFaceBitmap.Free;
      if Assigned(FBackBitmap) then FBackBitmap.Free;
      if Assigned(FaaBitmap) then FaaBitmap.Free;

      FBackBitmap:=NewDIBBitmap( Width * K, Height * K, pf24Bit);
      FFaceBitmap:=NewDIBBitmap(Width * K, Height * K, pf24Bit);
      FAABitmap:=NewDIBBitmap(Width, Height, pf24Bit);

      FBackbitmap.Canvas.Font.Assign(Font);
      FaaBitmap.Canvas.Font.Assign(Font);
      FFacebitmap.Canvas.Font.Assign(Font);
      FBackBitMap.Canvas.Font.FontHeight:= FaaBitmap.Canvas.Font.Fontheight * K;
      FBackBitMap.Canvas.Font.Fontquality := fqAntiAliased;
      RedrawScale;
   end;
end;

//****************************   ********************************

procedure THIAnalogGauge.DrawScale(Bitmap: PBitmap; K: Integer); //  
var   I, J, X, Y, N, M, W, H{, R}: Integer;
      Max, Min: Int64;
      A, C: Real;
      SI,CO, SI1,CO1:Extended;
      ARect: TRect;
      tl, tt, tw, th: Integer;
      SValue: string;
      Offset: Integer;
begin
      W := Bitmap.Width; H := Bitmap.Height; Max := FMaximum; Min := FMinimum;
      if fStyle in[ LeftStyle, RightStyle] then begin
         W:=KOL.Min(W,H);
         H:=KOL.Min(W,H);
      end;
      N := FNumMainTicks*5; M := FMargin * K; //R := FCircleRadius * K;
      with Bitmap{$ifndef F_P}^{$endif} do begin
      // ***************************** Out Frame **************************
         if ShowFrame in fFaceOptions then begin
            if Show3D in fFaceOptions then begin
               Canvas.Pen.PenWidth := 2*K;
               Canvas.Pen.Color := clBtnShadow;
               Canvas.MoveTo(W, 0); Canvas.LineTo(0, 0); Canvas.LineTo(0, H);
               Canvas.Pen.Color := clBtnHighlight;
               Canvas.LineTo(W, H); Canvas.LineTo(W, 0);
            end else begin
               Canvas.Pen.PenWidth := K;
               Canvas.Pen.Color := clBtnText;
               Canvas.Rectangle(0, 0, W, H);
            end;
         end;
         //************************* Out Margins **************************
         if ShowMargin in fFaceOptions then begin
            Canvas.Pen.Color := FMarginColor; Canvas.Pen.PenWidth := K;
            Canvas.Rectangle(M, M, W - M, H - M);
         end;
         //****************************************************************
         case fStyle of
            RightStyle: begin
               A := 0; C := W - M; X := W - M; Y := H - M;
               if fScaleAngle > 90 then fScaleAngle := 90; J := W - 2*M; end;
            LeftStyle: begin
               A := 90; C := M; X := M; Y := H - M;
               if fScaleAngle > 90 then fScaleAngle := 90; J := W - 2*M; end;
            else begin
               X := W div 2; A := (180 - fScaleAngle)/2; C := W/2;
               if fScaleAngle >= 180 then begin
                  J := (W - 2*M) div 2;
                  Y := H div 2;
               end else begin
                  J := Round(((W - 2*M - 2*FIndent*K)/2)/Cos(A*2*pi/360));
                  if J > H - 2*M then J := H - 2*M;
                  Y := (H - J) div 2 + J;
               end;
            end;
         end;{case}
         //******************************** Out Caption *******************
         if (ShowCaption in FFaceOptions) then begin
            Canvas.Brush.Color := FFaceColor;
            Canvas.Brush.BrushStyle := bsClear;
            Canvas.Font.Color := FCaptionColor;

            tw := Canvas.TextWidth(FCaption) + Canvas.TextExtent('M').cx;
            th := Canvas.TextHeight(FCaption);
            tl := Round(C - J/2 * Cos(A + FScaleAngle/2) *2*Pi/360) -  tw div 2;
            tt := Round(Y - J/2 * Sin((A + FScaleAngle/2)*2*Pi/360));
         
            if FStyle = CenterStyle then
               ARect := MakeRect(tl, tt, tw+tl, th+tt)
            else if FStyle = LeftStyle then
               ARect := MakeRect(tl+tw div 2+FMargin*K, tt, tw+tl+ tw div 2+FMargin*K, th+tt)
            else if FStyle = RightStyle then
               ARect := MakeRect(tl-tw div 2-FMargin*K, tt, tw+tl-tw div 2-FMargin*K, th+tt);
                        
            Canvas.RequiredState( HandleValid or FontValid or BrushValid or ChangingCanvas );
            DrawText(Canvas.Handle,PChar(FCaption),-1,ARect,DT_NOPREFIX or DT_VCENTER or DT_SINGLELINE or DT_CENTER);
         end;
         //*********************** Out MinMaxLines ************************
         Canvas.Pen.GeometricPen:=true;
         Canvas.Pen.PenEndCap:=pecFlat;//Square;
         Canvas.pen.penwidth:= 4 * k;
         if FGaugeMirror and not (FStyle in [CenterStyle]) then begin
               Max := FScaleValue - FMinimum;
               Min := FScaleValue - FMaximum;
         end;
         if (ShowIndicatorMax in FFaceOptions) then begin
            if not FGaugeMirror then
               Canvas.pen.color:=FMaxColor
            else if (FStyle in [RightStyle, LeftStyle]) and FGaugeMirror then 
               Canvas.pen.color:=FMinColor
            else               
               Canvas.pen.color:=FMaxColor;
            Sincos((A + FScaleAngle)*2*Pi/360,Si,Co);
            Sincos((A + Max*FScaleAngle/FScaleValue)*2*pi/360,Si1,Co1);
            Canvas.Arc(X - J, Y - J, X + J, Y + J,
                       Round(C - J * Co),
                       Round(Y - J * Si),
                       Round(C - J * CO1),
                       Round(Y - J * SI1))
         end;
         if (ShowIndicatorMid in FFaceOptions) and (FMinimum < FMaximum) then begin
            Canvas.pen.color:=FMidColor;
            Sincos((A + Max*FScaleAngle/FScaleValue)*2*pi/360,Si,Co);
            Sincos((A + Min*FScaleAngle/FScaleValue)*2*pi/360,Si1,Co1);
            Canvas.Arc(X - J, Y - J, X + J, Y + J,
                       Round(C - J * Co),
                       Round(Y - J * Si),
                       Round(C - J * Co1),
                       Round(Y - J * Si1))
         end;
         if (ShowIndicatorMin in FFaceOptions) then begin
            if not FGaugeMirror then
               Canvas.pen.color:=FMinColor
            else if (FStyle in [RightStyle, LeftStyle]) and FGaugeMirror then
               Canvas.pen.color:=FMaxColor
            else
               Canvas.pen.color:=FMinColor;                           
            Sincos((A + Min*FScaleAngle/FScaleValue)*2*Pi/360,si,co);
            Sincos(A*2*Pi/360,Si1,Co1);
            Canvas.Arc(X - J, Y - J, X + J, Y + J,
                       Round(C - J * Co),
                       Round(Y - J * Si),
                       Round(C - J * Co1),
                       Round(Y - J * Si1))
         end;
         Canvas.Font.Color := FValueColor;
         Canvas.Pen.Color := FTicksColor;
         Canvas.Pen.PenWidth := K;
         //************************* Out SubTicks *************************
         if ShowSubTicks in fFaceOptions then for I := 0 to N do begin
            Sincos((A+I*(FScaleAngle)/N)*2*pi/360,si,co);
            Canvas.MoveTo(Round(C-(J-FLengthSubTicks * K)*Co),
                          Round(Y-(J-FLengthSubTicks * K)*Si));
            Canvas.LineTo(Round(C-(J)*Co),
                          Round(Y-(J)*Si))
         end;
         //*********************** Out Main Ticks *************************
         for I := 0 to FNumMainTicks do begin
            if ShowMainTicks in fFaceOptions then begin
               SinCos((A+I*(FScaleAngle)/FNumMainTicks)*2*Pi/360,Si,Co);
               Canvas.MoveTo(Round(C-(J-FLengthMainTicks*K)*Co),
                             Round(Y-(J-FLengthMainTicks*K)*Si));
               Canvas.LineTo(Round(C-(J)*Co),
                             Round(Y-(J)*Si));
            end;
            //************************ Out Circles ***************************
            if ShowCircles in fFaceOptions then begin
               Canvas.Brush.Color := FCircleColor;
               sincos((A+I*(FScaleAngle)/FNumMainTicks)*2*Pi/360,Si,Co);
               Canvas.Ellipse(Round(C-(J*Co-fCircleRadius*K)),
                              Round(Y-(J*Si-fCircleRadius*K)),
                              Round(C-(J*Co+fCircleRadius*K)),
                              Round(Y-(J*Si+fCircleRadius*K)));
            end;
            // *********************** Out Values ****************************
            if ShowValues in fFaceOptions then begin
               Canvas.Brush.Color := FFaceColor;
               Canvas.Brush.BrushStyle := bsClear;
               Canvas.Font.Color := FValueColor;

               Offset := (fScaleValue - FMaximum) div (fScaleValue div FNumMainTicks); 

               if (not FGaugeMirror) and (FStyle in [CenterStyle]) then begin
                  if FTypeScale = scPlus then
                     SValue := Int2Str(i*fScaleValue div fNumMainTicks)
                  else if FTypeScale = scMinus then
                     SValue := Int2Str((fScaleValue - (FNumMainTicks-i+1)*fScaleValue) div fNumMainTicks)                  
                  else if FTypeScale = scMinusZPlus then
                     SValue := Int2Str((fScaleValue - (FNumMainTicks-i+1-Offset)*fScaleValue) div fNumMainTicks)
               end else if (FStyle in [RightStyle,LeftStyle]) and FGaugeMirror then begin
                  if FTypeScale = scPlus then
                     SValue := Int2Str((FNumMainTicks - i)*fScaleValue div fNumMainTicks)
                  else if FTypeScale = scMinus then
                     SValue := Int2Str((fScaleValue - (i+1)*fScaleValue) div fNumMainTicks)                   
                  else if FTypeScale = scMinusZPlus then
                     SValue := Int2Str((fScaleValue - (i+1-Offset)*fScaleValue) div fNumMainTicks)                   
               end else if (FStyle in [RightStyle,LeftStyle]) and not FGaugeMirror then begin
                  if FTypeScale = scPlus then
                     SValue := Int2Str(i*fScaleValue div fNumMainTicks)
                  else if FTypeScale = scMinus then
                     SValue := Int2Str((fScaleValue - (FNumMainTicks-i+1)*fScaleValue) div fNumMainTicks)
                  else if FTypeScale = scMinusZPlus then
                     SValue := Int2Str((fScaleValue - (FNumMainTicks-i+1-Offset)*fScaleValue) div fNumMainTicks)
               end else begin
                  if FTypeScale = scPlus then
                     SValue := Int2Str(i*fScaleValue div fNumMainTicks)
                  else if FTypeScale = scMinus then
                     SValue := Int2Str((fScaleValue - (FNumMainTicks-i+1)*fScaleValue) div fNumMainTicks)                  
                  else if FTypeScale = scMinusZPlus then
                     SValue := Int2Str((fScaleValue - (FNumMainTicks-i+1-Offset)*fScaleValue) div fNumMainTicks);                  
               end;
               tw := Canvas.TextWidth(SValue);
               th := Canvas.TextHeight(SValue);
               tl := Round(C-(J - fLengthMainTicks*K-5-I)*cos((A+i*(fScaleAngle)/fNumMainTicks)*2*pi/360));
               if FStyle in [LeftStyle] then
                  tl := tl - tw
               else if FStyle in [CenterStyle] then
                  tl := tl - tw div 2;
               tt := Round(Y-(J-fLengthMainTicks*K-5)*sin((A+i*(fScaleAngle)/fNumMainTicks)*2*pi/360));
               ARect := MakeRect(tl, tt, tw+tl, th+tt);
               Canvas.RequiredState( HandleValid or FontValid or BrushValid or ChangingCanvas );
               DrawText(Canvas.Handle,PChar(SValue),-1,ARect,DT_NOPREFIX or DT_VCENTER or DT_SINGLELINE or DT_CENTER);
            end;
         end;
      end;
end;

procedure THIAnalogGauge.DrawArrow(Bitmap: PBitmap; K: Integer); //  
var   J, X, Y, M, W, H, R: Integer;
      A, AA, C: Real;
      Si, Co: Extended;
      tl, tt, tw, th: Integer;
      ARect,BRect: TRect;
      Pos: Real;
begin
      if FPosition > FScaleValue then Pos := FScaleValue else Pos := FPosition;  
      M := FMargin * K; R := FCenterRadius * K; W := Bitmap.Width; H := Bitmap.Height;
      if fStyle in[ LeftStyle, RightStyle] then begin
         W:=kol.Min(W,H);
         H:=Kol.Min(W,H);
      end;
      with Bitmap{$ifndef F_P}^{$endif} do begin
         case FStyle of
            RightStyle: begin
               if FGaugeMirror then A := 90 else A := 0;
               AA := 0;
               C := W - M; X := W - M; Y := H - M;
               if FScaleAngle > 90 then FScaleAngle := 90; J := W - 2*M; end;
            LeftStyle: begin
               if FGaugeMirror then A := 180 else A := 90;
               AA := 90;
               C := M; X := M; Y := H - M;
              if FScaleAngle > 90 then FScaleAngle := 90; J := W - 2*M; end;
            else begin
               X := W div 2; A := (180 - fScaleAngle)/2; C := W/2;
               AA := A;
               if FScaleAngle >= 180 then begin
                  J := (W - 2*M) div 2;
                  Y := H div 2;
               end else begin
                  J := Round(((W - 2*M - 2*FIndent*K)/2)/Cos(A*2*pi/360));
                  if J > H - 2*M then J := H - 2*M;
                  Y := (H - J) div 2 + J;
               end;
            end;
         end;{case}
         //*********************** Led Max ********************************
         if (ShowLed in FFaceOptions) then begin
            tw := KOL.Min(H div 5,  40*K);
            th := KOL.Min(H div 10, 20*K) ;
            tl := Round(C - J/2 * Cos(AA + FScaleAngle/2) *2*Pi/360) -  tw div 2;
            tt := Round(Y - J/4 * Sin((AA + FScaleAngle/2)*2*Pi/360));

            if FStyle = CenterStyle then
               ARect := MakeRect(tl, tt, tw+tl, th+tt)
            else if FStyle = LeftStyle then
               ARect := MakeRect(tl+tw div 2+FMargin *K, tt-th div 2 , tw+tl+tw div 2+FMargin *K, tt+th div 2)
            else if FStyle = RightStyle then
               ARect := MakeRect(tl-tw div 2-FMargin *K, tt-th div 2, tl+tw div 2-FMargin *K, tt+th div 2);

            Canvas.Pen.PenWidth := K;
            Canvas.Brush.Color := GetShadeColor(FLedMaxColor, 120);               
            Canvas.Brush.BrushStyle := bsSolid;
            FillRect(Canvas.Handle,ARect,Canvas.Brush.Handle);

            if FPosition > FMaximum then begin
               BRect := ARect;
               dec(BRect.Bottom,K);
               dec(BRect.Right,K);
               Canvas.Brush.Color := FLedMaxColor;               
               Canvas.Brush.BrushStyle := bsSolid;
               FillRect(Canvas.Handle,BRect,Canvas.Brush.Handle);
               Canvas.Pen.Color := GetLightColor(FLedMaxColor, 75);
            end else
               Canvas.Pen.Color := GetShadeColor(FLedMaxColor, 50);

            SelectObject(Canvas.Handle,Canvas.Pen.Handle);
            MoveToEx(Canvas.Handle,ARect.Left+K*2,ARect.Top+K*4,nil);
            Windows.LineTo(Canvas.Handle,ARect.Left+K*8,ARect.Top+K*4);
            MoveToEx(Canvas.Handle,ARect.Left+K*2,ARect.Top+K*5,nil);
            Windows.LineTo(Canvas.Handle,ARect.Left+K*6,ARect.Top+K*5);
            MoveToEx(Canvas.Handle,ARect.Left+K*3,ARect.Top+K*8,nil);
            Windows.LineTo(Canvas.Handle,ARect.Left+K*3,ARect.Top+K*2);
            
            Canvas.Pen.Color := clGray;
            SelectObject(Canvas.Handle,Canvas.Pen.Handle);
            MoveToEx(Canvas.Handle,ARect.Left,ARect.Bottom,nil);
            Windows.LineTo(Canvas.Handle,ARect.Left,ARect.Top);
            Windows.LineTo(Canvas.Handle,ARect.Right,ARect.Top);
            Canvas.Pen.Color := clBtnShadow;
            SelectObject(Canvas.Handle,Canvas.Pen.Handle);
            MoveToEx(Canvas.Handle,ARect.Right,ARect.Top,nil);
            Windows.LineTo(Canvas.Handle,ARect.Right,ARect.Bottom);
            Windows.LineTo(Canvas.Handle,ARect.Left,ARect.Bottom);
           end;
         //****************************************************************
         
         Canvas.Pen.Penwidth := FArrowWidth * K; Canvas.Pen.Color := FArrowColor;
         Canvas.MoveTo(X, Y);
         if not FGaugeMirror then 
            Sincos((A + Pos*FScaleAngle/FScaleValue)*2*pi/360,si,co)
         else if (FStyle in [RightStyle, LeftStyle]) and FGaugeMirror then 
            Sincos((A - Pos*FScaleAngle/FScaleValue)*2*pi/360,si,co)
         else
            Sincos((A + Pos*FScaleAngle/FScaleValue)*2*pi/360,si,co);                     
         Canvas.LineTo(Round(C - J*Co),
                       Round(Y - J*Si));
         //********************** Out Center ******************************
         if ShowCenter in FFaceOptions then begin
            Canvas.Brush.Color := FCenterColor;
            Canvas.Ellipse(X - R, Y - R, X + R, Y + R);
         end;
      end;
end;

procedure THIAnalogGauge.RedrawArrow; //  
begin
   BitBlt(FFaceBitmap.Canvas.Handle, 0, 0, FBackBitmap.Width,
          FBackBitmap.Height, FBackBitmap.Canvas.Handle, 0, 0, SRCCOPY);
   DrawArrow(FFaceBitmap, GetAAMultipler);
   if FAntialiased <> aaNone then FastAntiAliasPicture;
   Control.Perform(WM_PAINT,-1,0);
end;


procedure THIAnalogGauge.RedrawScale; //  
begin
   FBackbitmap.Canvas.Brush.Color := FFaceColor;
   FBackbitmap.Canvas.Brush.BrushStyle := bsSolid;
   FBackbitmap.Canvas.Fillrect(FBackbitmap.Canvas.Cliprect);
   DrawScale(FBackBitmap, GetAAMultipler);
   RedrawArrow;
end;

procedure THIAnalogGauge.FastAntiAliasPicture; // 
var   x, y, cx, cy, cxi: Integer;
      totr, totg, totb: cardinal;
      Row1, Row2, Row3, Row4, DestRow: PRGBArray;
      i, k: Integer;
begin
      // For each row
      if not assigned(FFacebitmap) then exit;
      //  cx:=0;
      K := GetAAMultipler; Row2 := nil; Row3 := nil; Row4 := nil;
      for Y := 0 to FAABitmap.Height - 1 do begin
         // We compute samples of K x K pixels
         cy := y*K;
         // Get pointers to actual, previous and next rows in supersampled bitmap
         Row1 := FFaceBitmap.ScanLine[cy];
         if Row1 = nil then exit;
         if K > 1 then Row2 := FFaceBitmap.ScanLine[cy+1];
         if K > 2 then Row3 := FFaceBitmap.ScanLine[cy+2];
         if K > 3 then Row4 := FFaceBitmap.ScanLine[cy+3];
         // Get a pointer to destination row in output bitmap
         DestRow := FAABitmap.ScanLine[y];
         // For each column...
         for x := 0 to FAABitmap.Width - 1 do begin
            // We compute samples of 3 x 3 pixels
            cx := x*K;
            // Initialize result color
            totr := 0; totg := 0; totb := 0;
            if K > 3 then begin
               for i := 0 to 3 do begin
                  cxi := cx + i;
                  totr := totr + Row1[cxi].rgbtRed + Row2[cxi].rgbtRed + Row3[cxi].rgbtRed + Row4[cxi].rgbtRed;
                  totg := totg + Row1[cxi].rgbtGreen + Row2[cxi].rgbtGreen + Row3[cxi].rgbtGreen + Row4[cxi].rgbtGreen;
                  totb := totb + Row1[cxi].rgbtBlue + Row2[cxi].rgbtBlue + Row3[cxi].rgbtBlue + Row4[cxi].rgbtBlue;
               end;
               DestRow[x].rgbtRed := totr shr 4 ;//div 16;
               DestRow[x].rgbtGreen := totg shr 4 ;//16;
               DestRow[x].rgbtBlue := totb shr 4 ;//16;
            end else if K > 2 then begin
               for i := 0 to 2 do begin
                  cxi := cx + i;
                  totr := totr + Row1[cxi].rgbtRed + Row2[cxi].rgbtRed + Row3[cxi].rgbtRed;
                  totg := totg + Row1[cxi].rgbtGreen + Row2[cxi].rgbtGreen + Row3[cxi].rgbtGreen;
                  totb := totb + Row1[cxi].rgbtBlue + Row2[cxi].rgbtBlue + Row3[cxi].rgbtBlue;
               end;
               DestRow[x].rgbtRed := totr div 9;
               DestRow[x].rgbtGreen := totg div 9;
               DestRow[x].rgbtBlue := totb div 9;
            end else if K > 1 then begin
               for i := 0 to 1 do begin
                  cxi := cx + i;
                  totr := totr + Row1[cxi].rgbtRed + Row2[cxi].rgbtRed;
                  totg := totg + Row1[cxi].rgbtGreen + Row2[cxi].rgbtGreen;
                  totb := totb + Row1[cxi].rgbtBlue + Row2[cxi].rgbtBlue;
               end;
               DestRow[x].rgbtRed := totr  shr 2;
               DestRow[x].rgbtGreen := totg shr 2;
               DestRow[x].rgbtBlue := totb shr 2;
            end else begin
               DestRow[x].rgbtRed   := Row1[cx].rgbtRed;
               DestRow[x].rgbtGreen := Row1[cx].rgbtGreen;
               DestRow[x].rgbtBlue  := Row1[cx].rgbtBlue;
            end;
         end;
      end;
end;

function THIAnalogGauge.GetAAMultipler;
begin
   case FAntiAliased of
      aaBiline: Result := 2;
      aaTriline: Result := 3;
      aaQuadral: Result := 4;
      else Result := 1
   end;
end;

//******************************************************************************

procedure THIAnalogGauge._work_doArrowWidth;
begin
   FArrowWidth := ToInteger(_Data);
   if FArrowWidth < 1 then FArrowWidth := 1 else if FArrowWidth > 10 then FArrowWidth := 10;
   RedrawArrow;
   Control.Invalidate
end;

procedure THIAnalogGauge._work_doAngle;
begin
   FScaleAngle := ToInteger(_Data);
   if (FScaleAngle < 10) and (FScaleAngle > 360) then FScaleAngle := 10;   
   RedrawScale;
   Control.Invalidate;
end;

procedure THIAnalogGauge._work_doIndMinimum;
begin
   FMinimum := ToInteger(_Data);
   if (FMinimum < 1) or (FMinimum > FScaleValue) then FMinimum := 1;
   if FMinimum = FScaleValue then FMinimum := FScaleValue - 1;
   if FMinimum >= FMaximum then FMinimum := FMaximum - 1;
   RedrawScale;
   Control.Invalidate;
end;

procedure THIAnalogGauge._work_doIndMaximum;
begin
   FMaximum := ToInteger(_Data);
   if (FMaximum < 1) or (FMaximum > FScaleValue) then FMaximum := 1;
   if FMaximum = FScaleValue then FMaximum := FScaleValue - 1;
   if FMaximum <= FMinimum then FMaximum := FMinimum + 1;   
   RedrawScale;
   Control.Invalidate;
end;

procedure THIAnalogGauge._work_doScale;
begin
   FScaleValue := ToInteger(_Data);
   if FScaleValue <= FMaximum then FScaleValue := FMaximum + 1;
   RedrawScale;
   Control.Invalidate;
end;

procedure THIAnalogGauge.SetOption;
var   l:TFaceOptions;
begin
   l:= FFaceOptions;
   if OSet then include(l,Option)
   else exclude(l,Option);
   FFaceOptions:= l;
   RedrawScale;
end;

procedure THIAnalogGauge._work_doNumberMainTicks;
begin
   FNumMainTicks := ToInteger(_Data);
   if FNumMainTicks < 2 then FNumMainTicks := 2;  
   RedrawScale;
   Control.Invalidate;
end;

procedure THIAnalogGauge._work_doGaugeMirror;      begin FGaugeMirror := ReadBool(_Data);RedrawScale;Control.Invalidate;end;
procedure THIAnalogGauge._work_doGaugeCaption;     begin FCaption := ToString(_Data);RedrawScale;Control.Invalidate;end;
procedure THIAnalogGauge._work_doGaugeMargin;      begin FMargin := ToInteger(_Data);RedrawScale;Control.Invalidate;end;
procedure THIAnalogGauge._work_doCircleRadius;     begin FCircleRadius := ToInteger(_Data);RedrawScale;Control.Invalidate;end;
procedure THIAnalogGauge._work_doLengthMainTicks;  begin FLengthMainTicks := ToInteger(_Data);RedrawScale;Control.Invalidate;end;
procedure THIAnalogGauge._work_doLengthSubTicks;   begin FLengthSubTicks := ToInteger(_Data);RedrawScale;Control.Invalidate;end;
procedure THIAnalogGauge._work_doCenterRadius;     begin FCenterRadius := ToInteger(_Data);RedrawArrow;Control.Invalidate;end;
procedure THIAnalogGauge._work_doColorHover;       begin FColorHover := ReadBool(_Data);end;
procedure THIAnalogGauge._work_doTypeScale;        begin FTypeScale := TScale(ToInteger(_Data));RedrawScale;Control.Invalidate;end;
procedure THIAnalogGauge._work_doAntiAliased;      begin FAntiAliased := TAntialiased(ToInteger(_Data));Reinitialize;Control.Invalidate;end;
procedure THIAnalogGauge._work_doGaugeIndent;      begin FIndent := ToInteger(_Data);RedrawScale;Control.Invalidate;end;

procedure THIAnalogGauge._work_doShowMargin;       begin FShowMargin:= ReadBool(_Data);SetOption(FShowMargin,ShowMargin);Control.Invalidate;end;
procedure THIAnalogGauge._work_doShowCircles;      begin FShowCircles:= ReadBool(_Data);SetOption(FShowCircles,ShowCircles);Control.Invalidate;end;
procedure THIAnalogGauge._work_doShowMainTicks;    begin FShowMainTicks:= ReadBool(_Data);SetOption(FShowMainTicks,ShowMainTicks);Control.Invalidate;end;
procedure THIAnalogGauge._work_doShowSubTicks;     begin FShowSubTicks:= ReadBool(_Data);SetOption(FShowSubTicks,ShowSubTicks);Control.Invalidate;end;
procedure THIAnalogGauge._work_doShowIndicatorMin; begin FShowIndicatorMin:= ReadBool(_Data);SetOption(FShowIndicatorMin,ShowIndicatorMin);Control.Invalidate;end;
procedure THIAnalogGauge._work_doShowIndicatorMid; begin FShowIndicatorMid:= ReadBool(_Data);SetOption(FShowIndicatorMid,ShowIndicatorMid);Control.Invalidate;end;
procedure THIAnalogGauge._work_doShowIndicatorMax; begin FShowIndicatorMax:= ReadBool(_Data);SetOption(FShowIndicatorMax,ShowIndicatorMax);Control.Invalidate;end;
procedure THIAnalogGauge._work_doShowValues;       begin FShowValues:= ReadBool(_Data);SetOption(FShowValues,ShowValues);Control.Invalidate;end;
procedure THIAnalogGauge._work_doShowCenter;       begin FShowCenter:= ReadBool(_Data);SetOption(FShowCenter,ShowCenter);Control.Invalidate;end;
procedure THIAnalogGauge._work_doShowFrame;        begin FShowFrame:= ReadBool(_Data);SetOption(FShowFrame,ShowFrame);Control.Invalidate;end;
procedure THIAnalogGauge._work_doShow3D;           begin FShow3D:= ReadBool(_Data);SetOption(FShow3D,Show3D);Control.Invalidate;end;
procedure THIAnalogGauge._work_doShowCaption;      begin FShowCaption:= ReadBool(_Data);SetOption(FShowCaption,ShowCaption);Control.Invalidate;end;
procedure THIAnalogGauge._work_doShowLed;          begin FShowLed:= ReadBool(_Data);SetOption(FShowLed,ShowLed);Control.Invalidate;end;

procedure THIAnalogGauge._work_doMinColor;         begin FMinColor := ToInteger(_Data);RedrawScale;Control.Invalidate;end;
procedure THIAnalogGauge._work_doMidColor;         begin FMidColor := ToInteger(_Data);RedrawScale;Control.Invalidate;end;
procedure THIAnalogGauge._work_doMaxColor;         begin FMaxColor := ToInteger(_Data);RedrawScale;Control.Invalidate;end;
procedure THIAnalogGauge._work_doFaceColor;        begin FFaceColor := ToInteger(_Data);RedrawScale;Control.Invalidate;end;
procedure THIAnalogGauge._work_doFonHover;         begin FFonHover := ToInteger(_Data);end;
procedure THIAnalogGauge._work_doCaptionColor;     begin FCaptionColor := ToInteger(_Data);RedrawScale;Control.Invalidate;end;
procedure THIAnalogGauge._work_doTicksColor;       begin FTicksColor := ToInteger(_Data);RedrawScale;Control.Invalidate;end;
procedure THIAnalogGauge._work_doValueColor;       begin FValueColor := ToInteger(_Data);RedrawScale;Control.Invalidate;end;
procedure THIAnalogGauge._work_doCircleColor;      begin FCircleColor := ToInteger(_Data);RedrawScale;Control.Invalidate;end;
procedure THIAnalogGauge._work_doMarginColor;      begin FMarginColor := ToInteger(_Data);RedrawScale;Control.Invalidate;end;
procedure THIAnalogGauge._work_doArrowColor;       begin FArrowColor := ToInteger(_Data);RedrawArrow;Control.Invalidate;end;
procedure THIAnalogGauge._work_doCenterColor;      begin FCenterColor := ToInteger(_Data);RedrawArrow;Control.Invalidate;end;
procedure THIAnalogGauge._work_doLedMaxColor;      begin FLedMaxColor := ToInteger(_Data);RedrawArrow;Control.Invalidate;end;

procedure THIAnalogGauge._work_doFont;             begin if _IsFont(_Data) then SetFontRec(PFontRec(_Data.idata)^);Reinitialize;Control.Invalidate;end;

//******************************************************************************

end.
